import datetime

from django.urls import reverse

from web.forms import SubjectAddForm, SubjectEditForm
from web.models import StudySubject
from web.models.constants import SEX_CHOICES_MALE, SUBJECT_TYPE_CHOICES_CONTROL, SUBJECT_TYPE_CHOICES_PATIENT, \
    COUNTRY_AFGHANISTAN_ID, COUNTRY_OTHER_ID
from web.tests import LoggedInWithWorkerTestCase
from web.tests.functions import create_subject, create_visit, create_appointment, get_test_location
from web.views.notifications import get_today_midnight_date


class SubjectsViewTests(LoggedInWithWorkerTestCase):
    def setUp(self):
        super(SubjectsViewTests, self).setUp()

    def test_subjects_add(self):
        self.worker.save()

        response = self.client.get(reverse('web.views.subject_add'))
        self.assertEqual(response.status_code, 200)

    def test_render_subject_edit(self):
        subject = create_subject()

        response = self.client.get(reverse('web.views.subject_edit', kwargs={'id': subject.id}))
        self.assertEqual(response.status_code, 200)

    def test_render_subject_visit_details(self):
        subject = create_subject()
        visit = create_visit(subject)
        create_appointment(visit)

        response = self.client.get(reverse('web.views.subject_visit_details', kwargs={'id': subject.id}))
        self.assertEqual(response.status_code, 200)
        self.assertFalse("Add visit" in response.content)

    def test_render_subject_visit_details_without_visit(self):
        subject = create_subject()

        response = self.client.get(reverse('web.views.subject_visit_details', kwargs={'id': subject.id}))
        self.assertEqual(response.status_code, 200)
        self.assertTrue("Add visit" in response.content)

    def test_save_subject_edit_when_resigned_without_reason(self):
        subject = create_subject()
        form_subject = SubjectEditForm(instance=subject)
        form_data = {}
        for key, value in form_subject.initial.items():
            if value is not None:
                form_data[key] = value

        form_data['dead'] = "True"
        form_data['resigned'] = "True"
        response = self.client.post(
            reverse('web.views.subject_edit', kwargs={'id': subject.id}), data=form_data)

        self.assertEqual(response.status_code, 200)
        self.assertTrue("Resign reason cannot be empty" in response.content)

    def test_save_subject_edit(self):
        subject = create_subject()
        form_subject = SubjectEditForm(instance=subject)
        form_data = {}
        for key, value in form_subject.initial.items():
            if value is not None:
                form_data[key] = value

        form_data['dead'] = "True"
        form_data['resigned'] = "True"
        form_data['resign_reason'] = "doesn't want to participate"
        response = self.client.post(
            reverse('web.views.subject_edit', kwargs={'id': subject.id}), data=form_data)

        self.assertEqual(response.status_code, 302)
        updated_subject = StudySubject.objects.filter(id=subject.id)[0]
        self.assertTrue(updated_subject.dead)
        self.assertTrue(updated_subject.resigned)

    def test_subjects_add_2(self):

        form = SubjectAddForm(user=self.user)
        form_data = {}
        for key, value in form.initial.items():
            if value is not None:
                form_data[key] = value

        self.add_valid_form_data_for_subject_add(form_data)
        form_data["type"] = SUBJECT_TYPE_CHOICES_CONTROL
        form_data["default_location"] = get_test_location().id
        response = self.client.post(reverse('web.views.subject_add'), data=form_data)
        self.assertEqual(response.status_code, 302)
        response = self.client.get(response.url)
        self.assertContains(response, "Subject created")

        subject = StudySubject.objects.all()[0]
        self.assertEqual("L-001", subject.screening_number,
                         "prefix should start by L" +
                         " as default location prefix is not defined and subject type is control")

    @staticmethod
    def add_valid_form_data_for_subject_add(form_data):
        form_data["country"] = COUNTRY_AFGHANISTAN_ID
        form_data["first_name"] = "John"
        form_data["last_name"] = "Doe"
        form_data["sex"] = SEX_CHOICES_MALE
        form_data["type"] = SUBJECT_TYPE_CHOICES_PATIENT

    def test_subjects_add_patient(self):

        form = SubjectAddForm(user=self.user)
        form_data = {}
        for key, value in form.initial.items():
            if value is not None:
                form_data[key] = value

        self.add_valid_form_data_for_subject_add(form_data)
        form_data["default_location"] = get_test_location().id
        response = self.client.post(reverse('web.views.subject_add'), data=form_data)
        self.assertEqual(response.status_code, 302)
        response = self.client.get(response.url)
        self.assertContains(response, "Subject created")

        subject = StudySubject.objects.all()[0]
        self.assertEqual("P-001", subject.screening_number,
                         "prefix should start by P" +
                         " as default location prefix is not defined and subject type is patient")

    def test_subjects_add_invalid(self):

        form = SubjectAddForm(user=self.user)
        form_data = {}
        for key, value in form.initial.items():
            if value is not None:
                form_data[key] = value

        self.add_valid_form_data_for_subject_add(form_data)
        form_data["country"] = COUNTRY_OTHER_ID
        response = self.client.post(reverse('web.views.subject_add'), data=form_data)

        self.assertTrue("Invalid data" in response.content)

    def test_subjects_add_with_prefixed_location(self):

        form = SubjectAddForm(user=self.user)
        form_data = {}
        for key, value in form.initial.items():
            if value is not None:
                form_data[key] = value

        self.add_valid_form_data_for_subject_add(form_data)
        location = get_test_location()
        location.prefix = 'X'
        location.save()
        form_data["default_location"] = location.id
        response = self.client.post(reverse('web.views.subject_add'), data=form_data)
        self.assertEqual(response.status_code, 302)
        response = self.client.get(response.url)
        self.assertContains(response, "Subject created")

        subject = StudySubject.objects.all()[0]
        self.assertEqual("X-001", subject.screening_number,
                         "prefix should start by X as default location prefix is equal to 'X'")

    def test_render_subjects(self):
        create_subject()

        response = self.client.get(reverse('web.views.subjects'))
        self.assertEqual(response.status_code, 200)

    def test_render_subjects_with_no_visit(self):
        create_subject()

        response = self.client.get(reverse('web.views.subject_no_visits'))
        self.assertEqual(response.status_code, 200)

    def test_render_subjects_require_contact(self):
        subject = create_subject()
        subject.datetime_contact_reminder = get_today_midnight_date() + datetime.timedelta(days=-1)

        response = self.client.get(reverse('web.views.subject_require_contact'))
        self.assertEqual(response.status_code, 200)
